## no critic qw( RequireExistsError )
use Test::More tests => 55;
use strict;
use DE_EPAGES::Core::API::Constants;
use DE_EPAGES::Core::API::Error qw( CallerError CleanupOnError Error HandleError
                                    ExistsError GetError ResetError VarsOnError );
local $SIG{__DIE__} = \&DE_EPAGES::Core::API::Error::DieHandler;
#
# DO NOT REMOVE ANY OF THOSE LINES; THESE ARE REQUIRED FOR COUNTING THE CORRECT
# ERROR LINE (NO FUNNY STUFF)
#
#
sub MyFunction {
    my ( $Name, $Value ) = @_;

    # line 17
    Error( 'MISSING_PARAMETER', { Parameter => 'Name' } ) unless defined $Name;

    # line 20
    CallerError( 'VALUE_TOO_LONG', { Length => length $Value, MaxLength => 25 } ) if length $Value > 25;

    # line 23
    die 'Value too short' if length $Value < 3;

    return 'OK';
}

sub MyReEntryFunction {
    eval { MyFunction(); };
    Error(GetError());
}

sub MyReplaceDieHandlerFunction {
    local $SIG{__DIE__} = undef;
    eval { MyFunction('Native', '12'); };
    Error(GetError());
}

#ERRORDEFINITION
sub ErrorDefinition { return {
    'MISSING_PARAMETER' => 'Missing parameter #Parameter',
    'VALUE_TOO_LONG' => 'Value length #Length is larger than the allowed maximum #MaxLength'
}}
#ERRORDEFINITION END

my $DEBUG = 1;

# no error
my $Result = eval { MyFunction( 'Name', 'Value' ) };
ok( !ExistsError, 'eval without error' );
ok( 'OK' eq $Result, 'function result' );

# GetError
eval { $Result = MyFunction( undef, 'Value' ) };
ok( ExistsError(), 'eval with error' );
my $Error = GetError();
is( ref $Error, 'DE_EPAGES::Core::API::Error', 'Error is an Object reference' );
ok( $Error->isa('DE_EPAGES::Core::API::Error'), 'class of error object' );
is( $Error->code, 'MISSING_PARAMETER', 'error code' );
is( $Error->vars->{'Parameter'}, 'Name', 'error variable' );
ok( 1, 'dummy' );
ok( ExistsError(), 'GetError does not reset' );

# call stack
is( $Error->callers->[0]->[CALLER_LINE], 17, 'line number' );
is( $Error->callers->[0]->[CALLER_SUBNAME], 'DE_EPAGES::Core::API::Error::Error', 'Error() function name' );
is( $Error->callers->[1]->[CALLER_SUBNAME], 'main::MyFunction', 'caller function name' );

# ResetError
ResetError();
ok( !ExistsError(), 'ResetError' );

# shortMessage
ok( $Error->shortMessage eq 'Missing parameter Name', 'shortMessage' );

# longMessage
my $LongMessage = $Error->longMessage;
ok( $LongMessage =~ /^Missing parameter Name at \S*Error.t line 17.\n\n/, 'longMessage (subject)' );
ok( $LongMessage =~ /sub\ main\:\:MyFunction\ called\ at\ \S*Error\.t\ line\ \d+\./x, 'longMessage (caller)' );
ok( $LongMessage =~ /Vars{'Parameter'} = 'Name'/, 'longMessage (vars)' );

# die
eval { $Result = MyFunction( 'Name', 'V' ) };
ok( ExistsError(), 'native \'die\'' );
$Error = GetError();
is( $Error->code, 'NativeDie', 'error code for native \'die\'' );
my $msg = 'Value too short at '.__FILE__.' line 23.';
is( $Error->message, $msg, 'message for native \'die\'.');
is( $Error->shortMessage, 'Value too short', 'short message for native \'die\'.');

# without $SIG{__DIE__}
{
    my $Output;
    local $SIG{__DIE__} = sub { $Output = shift; };
    eval { $Result = MyFunction( undef, 'Value' ) };
    ok( $Output =~ /^MISSING_PARAMETER: Missing parameter Name at \S*Error\.t line 17\./, 'error without __DIE__ handler' );
    eval { $Result = MyFunction( 'Name', 'V' ) };
    ok( $Output =~ /^Value too short at \S*Error\.t line 23\./, 'die without __DIE__ handler' );
}

# re-entry stacks
eval { $Result = MyFunction( undef, 'Value' ) };
$Error = ExistsError() ? GetError() : undef;
eval {
    MyReEntryFunction();
};
$Error = GetError();
my $ReEntries = $Error->reEntries;
ok (scalar @$ReEntries == 1, 're-entry found');
is ($ReEntries->[0][0][CALLER_LINE], 30, 're-entry at line 30');

# local native die
eval { MyReplaceDieHandlerFunction(); };
$Error = GetError();
is ($Error->code, 'NativeDie', 'local native die');

# syntax error
eval "\n\n...;\nxxx;"; ## no critic (ProhibitStringyEval)
my $ExpectedMessage = $] < 5.012
    ? 'syntax error'
    : 'Bareword "xxx" not allowed while "strict subs" in use';
is( GetError()->shortMessage, $ExpectedMessage, 'syntax error: short message' );
like( GetError()->message, qr/^\Q$ExpectedMessage\E at /, 'syntax error: message' );

# inner error
eval {
    eval {
        MyFunction( undef, 'Value' );
    };
    ok( ExistsError(), 'caught inner error' );
    my $InnerError = GetError();
    Error( 'OuterError', { OuterVar => 'Value' }, $InnerError );
};
ok( ExistsError(), 'caught outer error' );
my $OuterError = GetError();
is( GetError->code, 'OuterError', 'outer error code' );
is( GetError->vars->{OuterVar}, 'Value', 'outer error vars' );
ok( GetError->innerLongMessage =~ /^Missing parameter Name/, 'innerLongMessage from error' );

# inner error message
eval {
    Error( 'OuterError', { OuterVar => 'Value' }, 'TestMessage' );
};
ok( GetError->innerLongMessage eq 'TestMessage', 'innerLongMessage from message' );

# HandleError
eval {
    eval {
        Error('Test_HandleError', { Var_1 => 'Value_1' });
    };
    HandleError(
        OnError => sub {
            my ($Err) = @_;
            is( $Err->code, 'Test_HandleError', 'HandleError - parameter of OnError' );
            $Err->addVars({ OnError => 1 });
            ok( !$Err->vars->{Finally}, 'HandleError - OnError called before Finally' );
        },
        Finally => sub {
            my ($Err) = @_;
            is( $Err->code, 'Test_HandleError', 'HandleError - parameter of Finally' );
            ok( $Err->vars->{OnError}, 'HandleError - Finally called after OnError' );
            $Err->addVars({ Finally => 1 });
        }
    );
};
ok( ExistsError(), 'HandleError - ExistsError' );
$Error = GetError();
is( $Error->code, 'Test_HandleError', 'HandleError - code' );
is( $Error->vars->{OnError}, 1, 'HandleError - OnError was called' );
is( $Error->vars->{Finally}, 1, 'HandleError - Finally was called' );

eval {
    eval {
        Error('Test_HandleError', { Var_1 => 'Value_1' });
    };
    HandleError( OnError => sub { return 'FIXED'; } );
};
ok( !ExistsError(), 'HandleError - OnError returned \'FIXED\'' );

# VarsOnError
eval {
    eval {
        Error('Test', { Var_1 => 'Value_1' });
    };
    VarsOnError({ Var_2 => 'Value_2' });
};
ok( ExistsError(), 'VarsOnError - ExistsError' );
is( GetError()->code, 'Test', 'VarsOnError - code' );
is( GetError()->vars->{Var_2}, 'Value_2', 'VarsOnError - vars' );

# CleanupOnError
eval {
    eval {
        Error('Test_Cleanup', { Var_1 => 'Value_1' });
    };
    CleanupOnError( sub {
        my ($Err) = @_;
        is( $Err->code, 'Test_Cleanup', 'CleanupOnError - parameter' );
        $Err->addVars({ Var_3 => 'Value_3' });
    });
};
ok( ExistsError(), 'CleanupOnError - ExistsError' );
is( GetError()->code, 'Test_Cleanup', 'CleanupOnError - code' );
is( GetError()->vars->{Var_3}, 'Value_3', 'CleanupOnError - vars' );

# CallerError
eval { MyFunction('Param1', 'This string is too long xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx') };
ok( ExistsError(), 'CallerError - ExistsError' );
$Error = GetError();
is( $Error->code, 'VALUE_TOO_LONG', 'CallerError - code' );
is( $Error->shortMessage, 'Value length 60 is larger than the allowed maximum 25 in call to main::MyFunction',
    'CallerError - shortMessage' );
like( $Error->message, qr/^Value length 60 is larger than the allowed maximum 25 in call to main::MyFunction at .*Error.t line \d+\./,
    'CallerError - message' );
like( $Error->longMessage, qr/^Value\ length\ 60\ is\ larger\ than\ the\ allowed\ maximum\ 25\ in\ call\ to\ main::MyFunction\ at\ \S*Error.t\ line\ \d+\.\n\n
                              sub\ \(eval\)\ called\ at\ \S*Error\.t\ line\ \d+\.\n
                              Code\ =\ VALUE_TOO_LONG\:[0-9A-F\-]+\n
                              Vars\{'Length'\}\ =\ '60'\n
                              Vars\{'MaxLength'\}\ =\ '25'\n
                              /xs,
    'CallerError - longMessage' );
