#!perl -w
use Test::More;
use strict;
use Config;
use Config::IniFiles;
use File::Temp qw(tempfile);

use DE_EPAGES::Core::API::File qw ( AbsolutePath DeleteDirectory DeleteFile
                                    CopyDirectory CopyFile CompareDirectories
                                    GetInfoDirectory GetFileContent WriteFile);
use DE_EPAGES::Core::API::Script qw( RunScript );

sub Main {

    plan tests => 7;

    my $PerlExe = $Config{'osname'} eq 'MSWin32' ? $^X : $ENV{'PERL'};

    my $PerlScript = AbsolutePath($0);
    $PerlScript =~ s,\\,/,g;
    $PerlScript =~ s/\.t$/\.pl/;
    $PerlScript =~ s,/t/,/Scripts/,;

    my $Path = $PerlScript;
    $Path =~ s,(/[^/]+){2}$,,;

    my (undef, $TmpFile) = tempfile();
    my (undef, $TmpFile2) = tempfile();

    my $Command = "$PerlExe $PerlScript -ini $TmpFile -text $Path";
    system("$Command");

    # test 1: does MD5Digest.pl works?
    cmp_ok($?, '==', 0, "system($Command)");

    DeleteDirectory("$Path.copy", {'recursive' => 1}) if -d "$Path.copy";

    CopyDirectory($Path, "$Path.copy", undef, { md5sum => $TmpFile });

    # test 2: after copying, directories should be equal
    ok(CompareDirectories($Path, "$Path.copy"), "CompareDirectories($Path, $Path.copy)");

    my $hInfo = GetInfoDirectory("$Path.copy", { recursive => 1 });

    my $ChangedFile;
    foreach (@{$hInfo->{Files}}) {
        if (-T $_) {
            $ChangedFile = $_; last;
        }
    }

    my $rContent = GetFileContent($ChangedFile);
    $$rContent =~ s/\r\n/chr(10)/ges;
    WriteFile($ChangedFile, $rContent);

    CopyDirectory($Path, "$Path.copy", undef, { md5sum => $TmpFile });

    $Command = "$PerlExe $PerlScript -ini $TmpFile2 -text $Path.copy";
    system("$Command");

    # test 3: does MD5Digest.pl still works?
    cmp_ok($?, '==', 0, "system($Command)");

    my $hInfo2 = GetInfoDirectory("$Path.copy", { recursive => 1 });
    my @Found;
    foreach (@{$hInfo2->{Files}}) {
        push @Found, $_ unless $_ ~~ @{$hInfo->{Files}};
    }
    my @Expected = ();

    # test 4: after copying, directories should be equal (ignore DOS-EOLs)
    is(@Found, @Expected, "(@Found) ~~ (@Expected)");

    $rContent = GetFileContent($ChangedFile);
    $$rContent .= "1;\n";
    WriteFile($ChangedFile, $rContent);

    CopyDirectory($Path, "$Path.copy", undef, { md5sum => $TmpFile } );

    $hInfo2 = GetInfoDirectory("$Path.copy", { recursive => 1 });
    @Found = ();
    foreach (@{$hInfo2->{Files}}) {
        push @Found, $_ unless $_ ~~ @{$hInfo->{Files}};
    }

    @Expected = ("$ChangedFile.latest");

    # test 5: after copying, directories differ: $ChangedFile.latest is new
    ok(@Found ~~ @Expected, "(@Found) ~~ (@Expected)");

    CopyFile($TmpFile2, "$TmpFile2.copy");

    $Command = "$PerlExe $PerlScript -ini $TmpFile2 -text $Path.copy";
    system("$Command");

    # test 6: does MD5Digest.pl still works?
    cmp_ok($?, '==', 0, "system($Command)");

    my (%IniHash, %IniHash2);
    tie %IniHash, 'Config::IniFiles', (-file => $TmpFile2)
        or die "$TmpFile2: Can't read file: $!";
    tie %IniHash2, 'Config::IniFiles', (-file => "$TmpFile2.copy")
        or die "$TmpFile2.copy: Can't read file: $!";

    @Found = ();
    foreach (keys %IniHash) {
        next if $_ eq '/';
        if (not defined $IniHash2{$_}{md5sum}) {
            push @Found, $_;
        } elsif (ref $IniHash{$_}{md5sum} ne ref $IniHash2{$_}{md5sum}) {
            push @Found, $_;
        } elsif (ref $IniHash{$_}{md5sum} eq 'ARRAY' and ref $IniHash2{$_}{md5sum} eq 'ARRAY') {
            push @Found, $_ unless @{$IniHash{$_}{md5sum}} ~~ @{$IniHash2{$_}{md5sum}};
        } else {
            push @Found, $_ unless $IniHash{$_}{md5sum} eq $IniHash2{$_}{md5sum};
        }
    }

    $ChangedFile =~ s,\Q$Path.copy\E/?,,;
    @Expected = sort ($ChangedFile, "$ChangedFile.latest");
    @Found = sort @Found;

    # test 7: what has changed in $TmpFile2 after rerun MD5Digest.pl?
    ok(@Found ~~ @Expected, "(@Found) ~~ (@Expected)");

    DeleteDirectory("$Path.copy", {'recursive' => 1}) if -d "$Path.copy";
    DeleteFile($TmpFile);
    DeleteFile($TmpFile2);
    DeleteFile("$TmpFile2.copy");
}

RunScript(Sub=>\&Main);
